﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.PTSP {
  [Item("Analytical Probabilistic Traveling Salesman Problem (PTSP)", "Represents a probabilistic traveling salesman problem where the expected tour length is calculated exactly.")]
  [Creatable(CreatableAttribute.Categories.CombinatorialProblems)]
  [StorableClass]
  public sealed class AnalyticalProbabilisticTravelingSalesmanProblem : ProbabilisticTravelingSalesmanProblem {

    [StorableConstructor]
    private AnalyticalProbabilisticTravelingSalesmanProblem(bool deserializing) : base(deserializing) { }
    private AnalyticalProbabilisticTravelingSalesmanProblem(AnalyticalProbabilisticTravelingSalesmanProblem original, Cloner cloner) : base(original, cloner) { }
    public AnalyticalProbabilisticTravelingSalesmanProblem() {
      Operators.Add(new BestPTSPSolutionAnalyzer());
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new AnalyticalProbabilisticTravelingSalesmanProblem(this, cloner);
    }

    public override double Evaluate(Permutation tour, IRandom random) {
      // Analytical evaluation
      double firstSum = 0;
      for (int i = 0; i < tour.Length - 1; i++) {
        for (int j = i + 1; j < tour.Length - 1; j++) {
          double sum1 = DistanceMatrix[tour[i], tour[j]] * Probabilities[tour[i]] * Probabilities[tour[j]];
          for (int k = i + 1; k < j; k++) {
            sum1 = sum1 * (1 - Probabilities[tour[k]]);
          }
          firstSum += sum1;
        }
      }
      double secondSum = 0;
      for (int j = 0; j < tour.Length - 1; j++) {
        for (int i = 0; i < j; i++) {
          double sum2 = DistanceMatrix[tour[j], tour[i]] * Probabilities[tour[i]] * Probabilities[tour[j]];
          for (int k = j + 1; k < tour.Length - 1; k++) {
            sum2 = sum2 * (1 - Probabilities[tour[k]]);
          }
          for (int k = 1; k < i; k++) {
            sum2 = sum2 * (1 - Probabilities[tour[k]]);
          }
          secondSum += sum2;
        }
      }
      return firstSum + secondSum;
    }

    public static double Evaluate(Permutation tour, DistanceMatrix distanceMatrix, DoubleArray probabilities) {
      // Analytical evaluation
      var firstSum = 0.0;
      for (var i = 0; i < tour.Length; i++) {
        for (var j = i + 1; j < tour.Length; j++) {
          var sum1 = distanceMatrix[tour[i], tour[j]] * probabilities[tour[i]] * probabilities[tour[j]];
          for (var k = i + 1; k < j; k++) {
            sum1 = sum1 * (1 - probabilities[tour[k]]);
          }
          firstSum += sum1;
        }
      }
      var secondSum = 0.0;
      for (var j = 0; j < tour.Length; j++) {
        for (var i = 0; i < j; i++) {
          var sum2 = distanceMatrix[tour[j], tour[i]] * probabilities[tour[i]] * probabilities[tour[j]];
          for (var k = j + 1; k < tour.Length; k++) {
            sum2 = sum2 * (1 - probabilities[tour[k]]);
          }
          for (var k = 0; k < i; k++) {
            sum2 = sum2 * (1 - probabilities[tour[k]]);
          }
          secondSum += sum2;
        }
      }
      return firstSum + secondSum;
    }
  }
}
